#ifndef ASSEMBLER_H
#define ASSEMBLER_H

#include "header.h"
//#include "contig.h"
#include "cluster.h"
#include "macros.h"
#include "random-access-vector.h"
#include "seeded-read.h"
#include "shorty-string.h"

class Contig;
class Hash;
class ReadProcessor;
//class TrieNode;

// Assembler for Harvard reads.
class Assembler {

 public:
  // Assembler(const vector<SeededRead>& psreads, ReadProcessor* rp, fstream* log,
  //	    fstream* contigs_out, fstream* geom_out);

  // This is used for debugging purpose. 'org_file_' is path of the file
  // containing the original sequence.
  Assembler(const vector<SeededRead>& psreads, ReadProcessor* rp,
	    fstream* log, fstream* contigs_out, fstream* geo_out,
	    fstream* geom_out, fstream* bambus_out);

  ~Assembler();

  // This is the main function to assemble the reads. It returns 'true' on success,
  // 'false' otherwise.
  bool Assemble();

  // Add an index of sread/hread to the corresponding index-list.
  inline void AddSRead(int i) { sreads_.push_back(i); }
  //  inline void AddHRead(int i) { hreads_.push_back(IntPair(i, -1)); }

  // Make clusters based on 'seeds'. Cluster created from left mars will be in
  // lcluster. Cluster created from right mars will be in rcluster.
  void MakeCluster(bool validate, const SStrIntPairVec& seeds,
		   Cluster* lcluster, Cluster* rcluster,
		   Contig* lcontig, Contig* rcontig);

  // Helper method for MakeCluster.
  void AddToCluster(const IntVec& ids, bool validate, int seed_id, int seed_pos,
		    mar_types mt, Cluster* cluster, Contig* contig);

  //  // Make a cluster based on 'seeds'. This is used for single strand mode.
  //void MakeCluster(const SStrIntPairVec& seeds, Cluster* lcluster,
  //	   Cluster* rcluster);

  // Take a cluster and create possible list of contigs.
  void MakeContigs(bool lcuster, const Cluster& cluster, const SeededRead& sread);

  bool Validate(const string& seed, const string& contig);

 private:
  // sreads_: index of the seeded reads, if the value is negative that mean it
  // is already used
  // hreads_: index of harvard reads, if the value is negative that mean it is
  // already used. The second int represents the current seed id for this read.
  // psreads_: points to the list of seeded reads
  // phreads_: points to the list of raw reads
  IntVec sreads_;
  //  IntPairVec hreads_;
  const vector<SeededRead>* psreads_;
  //  const vector<RawRead>* phreads_;

  map<int, int> reverse_comp_id_;
  
  // These are the newly created seeds.
  //list<SeededRead> newseeds_;

  // These are the cluster we formed with the help of the seeded reads.
  //vector<Cluster> clusters_;

  // List of contigs produced by the assembler. It is a list because we
  // will add/delete to/from this frequently while creating the contigs.
  //list<Contig> contigs_;

  // Pointer to a log file to record all events.
  fstream* log_;
  
  // Contigs are put in this file.
  fstream* contigs_out_;

  // Geometry information of contigs are put into this file.
  fstream* geom_out_;

  // This file will be used by bambus to generate scaffolds. Bambus expects
  // .contig format file (). See "http://amos.sourceforge.net/docs/bambus/
  // Manual.html#inputopt" for details.
  fstream* bambus_out_;
  fstream* geo_out_;

#ifdef DEBUG
  // This will be used to analyse/debug seed properties.
  fstream* seed_out_;
#endif  // DEBUG

  // Make clusters using the new generation seeds.
  void ProcessNewSeeds();

  //void Add2HashIndex(int val, int id, int offset, mar_types mt);

  // Output the unused reads at the end of assembling.
  //void DumpUnusedReads();

  //  void Add2Visualiser(const string& seed, const Cluster& lcluster,
  //		 const Cluster& rcluster);

  struct SeedSortByLen {
    bool operator ()(const SeededRead& a, const SeededRead& b) {
      return a.Str().size() < b.Str().size();
    }
  };

/*   struct SeedSortByGen { */
/*     bool operator ()(const SeededRead& a, const SeededRead& b) { */
/*       return a.Gen() < b.Gen(); */
/*     } */
/*   }; */

/*   struct SeedSortRandomly { */
/*     bool operator ()(const SeededRead& a, const SeededRead& b) { */
/*       return rand() % 2; */
/*     } */
/*   }; */

  // These are the newly created seeds.
#ifndef REUSE_CONTIG
  // priority_queue<SeededRead, deque<SeededRead>, SeedSortByLen> newseeds_;
#endif  // REUSE_CONTIG
#ifdef REUSE_CONTIG
  //priority_queue<SeededRead, deque<SeededRead>, SeedSortRandomly> newseeds_;
  //list<SeededRead> lseeds_, rseeds_;
  stack<SeededRead> lseeds_, rseeds_;
  //  RandomAccessVector<SeededRead> newseeds_;
#endif  // REUSE_CONTIG

  // Count the number of seeds. It is mainly used to assign an id to each seed.
  int seed_count_;
  int target_base_count_;
  int total_base_count_;
  static int leaf_contig_count_;

  ReadProcessor* read_processor_;
  //TrieNode* read_left_trie_;
  //TrieNode* read_right_trie_;
};

#endif  //  ASSEMBLER_H
