#include <iostream>
#include <fstream>

#include "cluster.h"
#include "contig.h"
#include "global.h"
#include "overlap-graph.h"
#include "read-processor.h"
#include "seeded-read.h"
#include "shorty-string.h"
#include "utils-string.h"
#include "utils-math.h"

OverlapGraph::OverlapGraph(bool right, const Cluster& cluster, 
			   const SeededRead* seed, const ReadProcessor* rp,
			   fstream* log, fstream* contigs_out)
  : right_(right), seed_(seed) {
  standalone_ = false;
  for (int i = 0; i < cluster.Indices().size(); ++i) {
    int id = cluster.GetIndex(i);
    //    int offset = cluster.GetOffset(i);
    int gen = cluster.GetGen(i);
    mar_types mt = cluster.GetMt(i);
    const ShortyString& mar = read_processor_->GetMar(id, mar_type_pair[mt]);
    // nodes_.push_back(ContigInfo(mar, id, -1, gen, mar_type_pair[mt]));
    AddNode(mar, id, -1, gen, mar_type_pair[mt]);
  }
  //AllocateData();
  read_processor_ = rp;
  log_ = log;
  contigs_out_ = contigs_out;
}

OverlapGraph::OverlapGraph() {
  standalone_ = true;
  ShortyString empty;
  seed_ = new SeededRead(empty, -1, -1);
}

void OverlapGraph::AllocateData() {
  adj_ = new IntTripleVec[nodes_.size()];
  curr_path_ = new IntPair[nodes_.size()];
  visited_ = new bool[nodes_.size()];
  path_valid_ = new bool[nodes_.size()];
  lstring_ = new ShortyString[nodes_.size()];
  lclen_ = new int[nodes_.size()];
  lpath_ = new IntPairVec[nodes_.size()]; 
}

OverlapGraph::~OverlapGraph() {
  vector<ContigInfo>().swap(nodes_);
  delete [] adj_;
  delete [] curr_path_;
  delete [] visited_;
  delete [] path_valid_;
  delete [] lstring_;
  delete [] lclen_;
  delete [] lpath_;
}

void OverlapGraph::AddNode(const ShortyString& mar, int id, int offset, int gen,
			   mar_types mt) {
  nodes_.push_back(ContigInfo(mar, id, offset, gen, mt));
}

void OverlapGraph::AddNode(const ShortyString& mar, int id, int gen) {
  nodes_.push_back(ContigInfo(mar, id, -1, gen, LMAR));
}

void OverlapGraph::BuildGraph() {
  for (int i = 0; i < nodes_.size(); ++i) {
    const string& str = nodes_[i].str_.Str();
    int val = UtilsString::String2Int(str, 0, HASH_LEN);
    buckets_[val].insert(i);
    for (int j = 1; j + HASH_LEN <= str.size(); ++j) {
      val = UtilsString::RightShiftStrVal(str[j - 1], str[j + HASH_LEN - 1],
					  val, HASH_LEN);
      buckets_[val].insert(i);
    }
    lstring_[i] = nodes_[i].str_;
    lclen_[i] = nodes_[i].size();
    lpath_[i].push_back(IntPair(i, 0));
  }
  for (int i = 0; i < (1 << (HASH_LEN << 1)); ++i) {
    set<int>::const_iterator ita = buckets_[i].begin();
    for (; ita != buckets_[i].end(); ++ita) {
      set<int>::const_iterator itb = buckets_[i].begin();
      for (; itb != buckets_[i].end(); ++itb) {
	if (*ita == *itb) continue;
	unique_pairs_.insert(IntPair(*ita, *itb));
      }
    }
  }
  set<IntPair, IntPairEq>::const_iterator ite = unique_pairs_.begin();
  for (; ite != unique_pairs_.end(); ++ite) {
    int i = ite->first;
    int j = ite->second;
    int o, s, offsetl;
    ShortyString merged;
    o = UtilsString::Overlap(nodes_[i].str_, nodes_[j].str_, &offsetl, &s,
			     &merged);
    if (o > 0) {
      adj_[i].push_back(IntTriple(j, offsetl, o));
    }
  }
  for (int i = 0; i < nodes_.size(); ++i) {
    sort(adj_[i].begin(), adj_[i].end(), SortAdj());
  }
}

void OverlapGraph::VisitGraph(vector<Contig>* contigs) {
  AllocateData();
  BuildGraph();
  for (int i = 0; i < nodes_.size(); ++i) {
    visited_[i] = false;
    path_valid_[i] = true;
  }
  for (int i = 0; i < nodes_.size(); ++i) {
    if (!visited_[i]) {
      Traverse(i, 0, nodes_[i].size(), 0);
    }
  }

  // Print the contigs.
  for (int i = 0; i < nodes_.size(); ++i) {
    if (lclen_[i] >= Global::DESIRED_CONTIG_LEN && path_valid_[i]) {
      MergeReads(lpath_[i]);
      Contig contig(right_, seed_->Id(), seed_->Size(), lstring_[i],
		    read_processor_);
      for (int j = 0; j < lpath_[i].size(); ++j) {
	const ContigInfo& node = nodes_[lpath_[i][j].first];
	contig.AddRead(node.id_, lpath_[i][j].second, node.gen_, node.mt_);
      }
      contigs->push_back(contig);
    }
  }
}

void OverlapGraph::Traverse(int root, int depth, int elength, int poffset) {
  visited_[root] = true;
  curr_path_[depth] = IntPair(root, poffset);
  bool flag = false;
  for (int i = 0; i < adj_[root].size(); ++i) {
    int j = adj_[root][i].first;
    if (visited_[j]) {
      
      // adj_[root][i] was visited previously. So, if we add that to the current
      // path, it will be bigger.
      int add = (nodes_[root].size() - adj_[root][i].second) > lclen_[j] ? 0
	: (lclen_[j] - nodes_[root].size() + adj_[root][i].second);
      int newlen = elength - poffset < nodes_[root].size() + add ? 
	(poffset + nodes_[root].size() + add) : elength;
      int start = curr_path_[0].first;
      if (newlen >= elength && elength >= lclen_[start]) {
	path_valid_[j] = false;
	lclen_[start] = newlen;
	IntPairVec tpath; // Need this to avoid a cycle.
	//lpath_[start].clear();
	for (int k = 0; k <= depth; ++k) {
	  tpath.push_back(curr_path_[k]);
	}
	for (int k = 0; k < lpath_[j].size(); ++k) {
	  tpath.push_back(IntPair(lpath_[j][k].first,
				  adj_[root][i].second + lpath_[j][k].second + poffset));
	}
	lpath_[start] = tpath;
      }
      continue;
    }
    flag = true;
    int add = (nodes_[root].size() - adj_[root][i].second) > nodes_[j].size() ? 0
      : (nodes_[j].size() - nodes_[root].size() +
	 adj_[root][i].second);
    int newlen = elength - poffset < nodes_[root].size() + add ? 
      (poffset + nodes_[root].size() + add) : elength;
    Traverse(adj_[root][i].first, depth + 1, newlen, poffset +
	     adj_[root][i].second);
  }
  if (!flag) {
    int start = curr_path_[0].first;
    if (elength > lclen_[start]) {
      lclen_[start] = elength;
       lpath_[start].clear();
      for (int i = 0; i <= depth; ++i) {
	lpath_[start].push_back(curr_path_[i]);
      }
    }
  }
}

void OverlapGraph::MergeReads(const IntPairVec& reads) {
  assert(reads.size() > 0);
  ShortyString ret = nodes_[reads[0].first].str_;
  for (int i = 1; i < reads.size(); ++i) {
    ret = UtilsString::OnlyMergeWithoutDel(ret, nodes_[reads[i].first].str_,
    					   reads[i].second);
  }
  if (ret.size() > lstring_[reads[0].first].size()) {
    lstring_[reads[0].first] = ret;
  }
}

// void OverlapGraph::ReUseReads() {
//   for (int i = 0; i < index_.size(); ++i) {
//     if (index_[i].first == i) {
//       const IntPairVec& ids = nodes_[i].Ids();
//       if (ids.size() == 1) {
// 	read_processor_->Reset(ids[0].first);
// 	index_[i].first = -1;
//       }
//     }
//   }
// }

// void OverlapGraph::ReUseReads(int x) {
//   const IntPairVec& ids = nodes_[x].Ids();
//   for (int i = 0; i < ids.size(); ++i) {
//     read_processor_->Reset(ids[i].first);
//   }
//   index_[x].first = -1;
// }
