// This is a driver program to generate statistics using MUMmer.
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <numeric>
#include <cstdlib>

using namespace std;

const int INF = 1000000000;

vector<double> acc;
vector<int*> cov, covrc;

map<string, int> contig_pos;

typedef struct contig_info {
  bool rc;
  int s, e, l;
  double score;
  string label;
};

int total;

const int acc_val_size = 5;
double acc_val[] = {90, 95, 97, 99, 99.9};

vector<contig_info> contig_infos;
vector<double> cov_stat_val;
vector<double> usg_stat_val;
vector<int> contig_len_bins[acc_val_size];

void gap_analysis(const int* c, int size) {
  int head_gap = 0;
  for (int i = 0; i < size && !c[i]; ++i, ++head_gap);
  int tail_gap = 0;
  for (int i = size - 1; i >= 0 && !c[i]; --i, ++tail_gap);

  // Other gaps.
  vector<int> gaps;
  int curr = 0;
  for (int i = head_gap; i < size - tail_gap; ++i) {
    if (c[i]) {
      if (curr) gaps.push_back(curr);
      curr = 0;
      continue;
    }
    ++curr;
  }
  int maxg = 0;
  int ming = size;
  double avg = 0;
  for (int i = 0; i < gaps.size(); ++i) {
    maxg = max(maxg, gaps[i]);
    ming = min(ming, gaps[i]);
    avg += gaps[i];
  }
  avg /= gaps.size();
  cout << head_gap << "(" << 100.0 * head_gap / size << ")\t";
  cout << tail_gap << "(" << 100.0 * tail_gap / size << ")\t";
  cout << gaps.size() << "\t";
  cout << avg << "(" << 100.0 * avg / size << ")\t";
  cout << maxg << "(" << 100.0 * maxg / size << ")\t";
  cout << ming << "(" << 100.0 * ming / size << ")" << endl;
}

void contig_distribution(int min_len, int max_len, int ref_size) {
  cout << "--------------contig info----------------\n";
  cout << "total " << contig_infos.size() << " min-allowed " << min_len
       << " max-allowed " << max_len;
  int maxl = -INF;
  int minl = INF;
  double avgl = 0;
  double cov;
  vector<int> lens, votes;
  lens.push_back(50); votes.push_back(0);
  lens.push_back(100); votes.push_back(0);
  lens.push_back(200); votes.push_back(0);
  lens.push_back(300); votes.push_back(0);
  lens.push_back(500); votes.push_back(0);
  lens.push_back(700); votes.push_back(0);
  lens.push_back(1000); votes.push_back(0);
  lens.push_back(1500); votes.push_back(0);
  lens.push_back(2000); votes.push_back(0);
  lens.push_back(2500); votes.push_back(0);
  lens.push_back(3000); votes.push_back(0);
  lens.push_back(3500); votes.push_back(0);
  lens.push_back(4000); votes.push_back(0);
  lens.push_back(4500); votes.push_back(0);
  lens.push_back(5000); votes.push_back(0);
  lens.push_back(10000); votes.push_back(0);
  lens.push_back(15000); votes.push_back(0);
  lens.push_back(20000); votes.push_back(0);
  for (int i = 0; i < contig_infos.size(); ++i) {
    avgl += contig_infos[i].l;
    maxl = max(maxl, contig_infos[i].l);
    minl = min(minl, contig_infos[i].l);
    for (int j = 0; j < lens.size(); ++j) {
      if (contig_infos[i].l >= lens[j]) {
	++votes[j];
      }
    }
  }
  cov = avgl / ref_size;
  avgl /= contig_infos.size();
  cout << " min " << minl << " max " << maxl << " avg " << avgl << " thck "
       << cov << endl;
  for (int i = 0; i < lens.size(); ++i) {
    cout << '[' << lens[i] << "-" << votes[i] << ']';
  }
  cout << endl; 
  cout << "------------------------------------------\n";
}

double get_coverage(int* c, int size) {
  int count = 0;
  for (int i = 0; i < size; ++i) {
    if (c[i]) ++count;
  }
  return 100.0 * count / size;
}

double print_layout(const int* c, int size) {
  int start = -1;
  int gstart = -1;
  int maxc = 0;
  for (int i = 0; i < size; ++i) {
    if (c[i]) {  // Start or end?
      if (start == -1) start = i;
      if (gstart != -1) {
	cout << "(" << i - gstart << ")";
	gstart = -1;
      }
    } else {
      if (start != -1) {
	cout << "[" << start << "-" << i - 1 << "]";
	maxc = max(maxc, i - start);
	start = -1;
      }
      if (gstart == -1) gstart = i;
    }
  }
  if (start != -1) {
    cout << "[" << start << "-" << size - 1 << "]";
    maxc = max(maxc, size - start);
  }
  if (gstart != -1) {
    cout << "(" << size - gstart << ")";
  }
  cout << endl;
  cout << "maximum possible contig size: " << maxc << endl;
}

double get_contig_ratio(bool rc, double a) {
  int count = 0;
  for (int i = 0; i < contig_infos.size(); ++i) {
    if (contig_infos[i].rc == rc && contig_infos[i].score >= a) {
      ++count;
    }
  }
  return 100.0 * count / total;
}

double permutation_analysis(const vector<int>& p) {
  // Increasing order first.
  vector<int> inc(p.size());
  inc[0] = 1;
  for (int i = 1; i < p.size(); ++i) {
    inc[i] = 1;
    for (int j = 0; j < i; ++j) {
      if (p[i] >= p[j]) {
	inc[i] = max(inc[i], inc[j] + 1);
      }
    }
  }
  vector<int> dec(p.size());
  dec[0] = 1;
  for (int i = 1; i < p.size(); ++i) {
    dec[i] = 1;
    for (int j = 0; j < i; ++j) {
      if (p[i] <= p[j]) {
	dec[i] = max(dec[i], dec[j] + 1);
      }
    }
  }
  int mi = -INF;
  int md = -INF;
  for (int i = 0; i < p.size(); ++i) {
    mi = max(mi, inc[i]);
    md = max(md, dec[i]);
  }
  //  delete(inc);
  //delete(dec);
  //  cout << "mi " << mi << " " << "md " << md << endl;
  return 100.0 * max(mi, md) / p.size();
}

void Test() {
  vector<int> vi;
  vi.push_back(0);
  vi.push_back(1);
  vi.push_back(2);
  vi.push_back(3);
  vi.push_back(0);
  vi.push_back(1);
  vi.push_back(2);
  vi.push_back(3);
  permutation_analysis(vi);
}

int main(int argc, char** argv) {
  if (argc != 4) {
  //  if (argc != 5) {
    //cout << "params: ref-len min-allowed max-allowed chain-file\n";
    cout << "params: ref-len min-allowed chain-file\n";
    exit(0);
  }
  fstream cov_stat("out/cov_stat", ios::out | ios::app);
  fstream usg_stat("out/usg_stat", ios::out | ios::app);
  total = 0;
  int org_len = atoi(argv[1]);
  int min_contig_len = atoi(argv[2]);
  //  int max_contig_len = atoi(argv[3]);
  int max_contig_len = INF;
  cout << org_len << "*****" << min_contig_len << endl; 

  // The accuracy levels we are interested in.
  for (int i = 0; i < acc_val_size; ++i) {
    acc.push_back(acc_val[i]);
    cov.push_back(new int [org_len]);
    fill(cov[i], cov[i] + org_len, 0);
    covrc.push_back(new int [org_len]);
    fill(covrc[i], covrc[i] + org_len, 0);
  }
  string str;

  // Read the first line.
  cin >> str;
  cout << str << ' ';
  cin >> str;
  cout << str << endl;

  // Read the second line and do nothing.
  cin >> str;

  // Read the third line and do nothing.
  for (int i = 0; i < 19; ++i) {
    cin >> str;
  }

  // Process all the remaining entries in the file.
  int s1, e1, s2, e2, l1, l2, lr, lq;
  double idy, covr, covq;
  string tag, label;
  while (cin >> s1) {
    cin >> e1;
    cin >> s2;
    cin >> e2;
    cin >> l1;
    cin >> l2;
    cin >> idy;
    cin >> lr;
    cin >> lq;
    cin >> covr;
    cin >> covq;
    cin >> tag;
    cin >> label;
    if (lq < min_contig_len || lq > max_contig_len) continue;
    double score = idy - 100 + covq;
    int pos = contig_pos[label];
    if (pos == 0) {  // not seen before.
      contig_pos[label] = ++total;
      pos = total - 1;
    } else {
      --pos;
    }
    if (pos == contig_infos.size()) {  // new entry.
      contig_info t;
      t.rc = (tag[0] == '+' ? false : true); 
      t.s = s1;
      t.e = e1;
      t.l = lq;
      t.score = score;
      t.label = label;
      contig_infos.push_back(t);
    } else {
      if (score > contig_infos[pos].score) {
	contig_info t;
	t.rc = (tag[0] == '+' ? false : true); 
	t.s = s1;
	t.e = e1;
	t.l = lq;
	t.score = score;
	t.label = label;
	contig_infos[pos] = t;
      }
    }
  }
  for (int i = 0; i < contig_infos.size(); ++i) {
    for (int j = 0; j < acc_val_size; ++j) {
      if (contig_infos[i].score >= acc_val[j]) {
	contig_len_bins[j].push_back(contig_infos[i].l);
	for (int k = contig_infos[i].s; k <= contig_infos[i].e; ++k) {
	  if (contig_infos[i].rc) {
	    ++covrc[j][k];
	    ++cov[j][org_len - k];
	  } else {
	    ++cov[j][k];
	    ++covrc[j][org_len - k];
	  }
	}
      }
    }
  }
  //  cout << "min read len: " << min_contig_len << endl;
  cov_stat << min_contig_len;
  usg_stat << min_contig_len;
  //cout << "total # of contigs: " << total << endl;
  int total_bases = 0;
  for (int i = 0; i < contig_infos.size(); ++i) {
    total_bases += contig_infos[i].l;
  }
  //cout << "total coverage: " << (double) total_bases / org_len << endl;
  contig_distribution(min_contig_len, max_contig_len, org_len);
  cout.precision(4);
  cout << "+/-";
  for (int i = 0; i < acc_val_size; ++i) {
    cout << "\t" << acc_val[i] << "(%)";
  }
  cout << "\n----------------------------------------------" << endl;
  cout << "+";
  for (int i = 0; i < acc_val_size; ++i) {
    cout << "\t" << get_coverage(cov[i], org_len) << '(' << 
      get_contig_ratio(false, acc_val[i]) << ')';
    //cov_stat << "\t" << get_coverage(cov[i], org_len);
    cov_stat_val.push_back(get_coverage(cov[i], org_len));
  }
  cout << endl;
  //  cov_stat << endl;
  cout << "-";
  for (int i = 0; i < acc_val_size; ++i) {
    cout << "\t" << get_coverage(covrc[i], org_len) << '(' << 
      get_contig_ratio(true, acc_val[i]) << ')';
    cov_stat_val[i] = max(cov_stat_val[i], get_coverage(covrc[i], org_len));
    cov_stat << "\t" << cov_stat_val[i];
  }
  cout << "\n----------------------------------------------" << endl;
  cout << '\t';
  cov_stat << endl;
  for (int i = 0; i < acc_val_size; ++i) {
    cout << '\t' << get_contig_ratio(false, acc_val[i]) + get_contig_ratio(true, acc_val[i]);
    usg_stat << '\t' << get_contig_ratio(false, acc_val[i]) + get_contig_ratio(true, acc_val[i]);
  }
  cout << "\n----------------------------------------------" << endl;
  usg_stat << endl;
 
  // Get a single score for the assembler.
  double single_accuracy = 0;
  int total_contig_len = 0;
  for (int i = 0; i < contig_infos.size(); ++i) {
    single_accuracy += contig_infos[i].score * contig_infos[i].l;
    total_contig_len += contig_infos[i].l;
  }
  single_accuracy /= total_contig_len;
  cout << "whole assembly score: " << single_accuracy << "\n--------------\n";
  cout << "acc\t" << "headg\t" << "tailg\t" << "#\t" << "avg\t" << "maxg\t"
       << "ming\n"; 
  cout << "------------gap(+)------------\n";
  for (int i = 0; i < acc_val_size; ++i) {
    cout << acc_val[i] << "\t";
    gap_analysis(cov[i], org_len);
  }
  cout << "------------gap(-)------------\n";
  for (int i = 0; i < acc_val_size; ++i) {
    cout << acc_val[i] << "\t";
    gap_analysis(covrc[i], org_len);
  }
  ofstream avg_out("out/avg.tmp");
  for (int i = 0; i < acc_val_size; ++i) {
    avg_out << accumulate(contig_len_bins[i].begin(),
			  contig_len_bins[i].end(), 0) /
      (double)contig_len_bins[i].size() << " ";
  }
  ofstream contig_95("out/contigs.95.m");
  ofstream contig_97("out/contigs.97.m");
  ofstream contig_99("out/contigs.99.m");
  ofstream contig_99_9("out/contigs.99_9.m");
  for (int i = 0; i < contig_infos.size(); ++i) {
    double score = contig_infos[i].score;
    if (score >= 95.0) {
      contig_95 << contig_infos[i].l << endl;
    }
    if (score >= 97.0) {
      contig_97 << contig_infos[i].l << endl;
    }
    if (score >= 99.0) {
      contig_99 << contig_infos[i].l << endl;
    }
    if (score >= 99.9) {
      contig_99_9 << contig_infos[i].l << endl;
    }
  }
  contig_95.close();
  contig_97.close();
  contig_99.close();
  contig_99_9.close();
  ofstream contig_summary_out("out/contig-summary.tmp");
  for (int i = 0; i < contig_infos.size(); ++i) {
    contig_summary_out << contig_infos[i].label << " (" << contig_infos[i].l 
		       << ")[" << contig_infos[i].s << "-" << contig_infos[i].e
		       << "]" << contig_infos[i].score << endl;
  }
  avg_out << endl;
  avg_out.close();
  cov_stat.close();
  usg_stat.close();
  contig_summary_out.close();
  //Test();

  //   cout << "following contigs were processed:\n";
  //   for (int i = 0; i < contig_infos.size(); ++i) {
  //     cout << " " << contig_infos[i].label;
  //   }
  //   cout << endl;
  
  // Chain analysis.
  fstream fchain(argv[4], ios::in);
  int chain_len;
  vector<pair<double, int> > chain_scores;
  while (fchain >> chain_len) {
    vector<int> chain_ids;
    string chain_str_id;
    for (int i = 0; i < chain_len; ++i) {
      fchain >> chain_str_id;
      int pos = contig_pos[chain_str_id];
      if (pos) chain_ids.push_back(contig_infos[pos - 1].s);
      else chain_ids.push_back(-1 * (rand() % INF));
    }
    chain_scores.push_back(pair<double, int>(permutation_analysis(chain_ids), chain_len));
  }
  //  cout << "------------chains------------\n";
  //cout << "score\tlength\n";
  //for (int i = 0; i < chain_scores.size(); ++i) {
  //  cout << chain_scores[i].first << "\t" << chain_scores[i].second << endl;
  // }
  cout << "-----------layout(+)-----------\n";
  for (int i = 0; i < acc_val_size; ++i) {
    cout << ">" << acc_val[i] << "%\n";
    print_layout(cov[i], org_len);
  }
  cout << "-----------layout(-)-----------\n";
  for (int i = 0; i < acc_val_size; ++i) {
    cout << ">" << acc_val[i] << "%\n";
    print_layout(covrc[i], org_len);
  }
  return 0;
}
